import type { DbDict } from 'kmore-types';
import type { Knex } from 'knex';
import { KmoreBase } from './base.js';
import { DbQueryBuilder, KmoreQueryBuilder } from './builder.types.js';
import { postProcessResponse } from './helper.js';
import { CaseType, EventCallbacks, KmoreTransaction, KmoreTransactionConfig, KnexConfig, QueryContext, TrxIdQueryMap } from './types.js';
export declare class Kmore<D = any, Context = any> extends KmoreBase<Context> {
    /**
     * Original table names, without case convertion.
     */
    readonly refTables: DbQueryBuilder<Context, D, 'ref_', CaseType.none>;
    /**
     * Create a table reference function property with camel case convertion.
     */
    readonly camelTables: DbQueryBuilder<Context, D, 'ref_', CaseType.camel>;
    /**
     * Create a table reference function property with snake case convertion.
     */
    readonly snakeTables: DbQueryBuilder<Context, D, 'ref_', CaseType.snake>;
    /**
    * Generics parameter, do NOT access as variable!
    * Use under typescript development only.
    *
    * @example ```ts
    * const km = kmoreFactore<Db>({})
    * type DbModel = typeof km.DbModel
    * type Uid = DbModel['tb_user']['uid']  // equal to number
    * ```
    */
    readonly DbModel: D;
    /**
    * Generics parameter, do NOT access as variable!
    * Use under typescript development only.
    *
    * @example ```ts
    * const km = kmoreFactore<Db>({})
    * type Dict = typeof kmore['Dict']
    * type Uid = Dict['columns']['tb_user']['alias']['tbUserUid'] // equal to literal 'tb_user.uid'
    * ```
    */
    readonly Dict: DbDict<D>;
    readonly postProcessResponseSet: Set<typeof postProcessResponse>;
    readonly trxActionOnEnd: KmoreTransactionConfig['trxActionOnEnd'];
    /**
     * kmoreTrxId => trx
     */
    readonly trxMap: Map<symbol, KmoreTransaction>;
    /**
     * kmoreTrxId => Set<kmoreQueryId>
     */
    readonly trxIdQueryMap: TrxIdQueryMap;
    /**
     * context => Set<kmoreTrxId>
     */
    readonly ctxTrxIdMap: WeakMap<object, Set<symbol>>;
    readonly config: KnexConfig;
    readonly dict: unknown extends D ? undefined : DbDict<D>;
    readonly dbId: string;
    readonly dbh: Knex;
    readonly instanceId: string | symbol;
    readonly eventCallbacks: EventCallbacks<Context> | undefined;
    readonly wrapIdentifierCaseConvert: CaseType;
    constructor(options: KmoreFactoryOpts<D, Context>);
    /**
     * Start a transaction.
     */
    transaction(config?: KmoreTransactionConfig): Promise<KmoreTransaction>;
    getTrxByKmoreQueryId(kmoreQueryId: symbol): KmoreTransaction | undefined;
    getTrxByKmoreTrxId(id: symbol): KmoreTransaction | undefined;
    setCtxTrxIdMap(ctx: unknown, kmoreTrxId: symbol): void;
    getTrxSetByCtx(ctx: unknown): Set<KmoreTransaction>;
    finishTransaction(trx: KmoreTransaction | undefined): Promise<void>;
    finishTransactionByQueryId(kmoreQueryId: symbol): Promise<void>;
    destroy(): Promise<void>;
    raw<T = unknown>(...args: unknown[]): Promise<T>;
    builderClone(builder: KmoreQueryBuilder): KmoreQueryBuilder;
    protected postProcessResponse(result: any, queryContext?: QueryContext): unknown;
}
export interface KmoreFactoryOpts<D, Ctx = unknown> {
    config: KnexConfig;
    dict?: DbDict<D>;
    instanceId?: string | symbol;
    dbh?: Knex;
    dbId?: string;
    /**
     * @docs https://knexjs.org/guide/interfaces.html#start
     * @docs https://knexjs.org/guide/interfaces.html#query
     * @docs https://knexjs.org/guide/interfaces.html#query-response
     */
    eventCallbacks?: EventCallbacks<Ctx> | undefined;
    /**
     * Table identifier case convertion,
     * If not CaseType.none, will ignore value of `KnexConfig['wrapIdentifier']`
     * @default CaseType.snake
     * @docs https://knexjs.org/guide/#wrapidentifier
     */
    wrapIdentifierCaseConvert?: CaseType | undefined;
    /**
     * Atuo trsaction action (rollback|commit|none) on error (Rejection or Exception),
     * @CAUTION **Will always rollback if query error in database even though this value set to 'commit'**
     * @default rollback
     */
    trxActionOnEnd?: KmoreTransactionConfig['trxActionOnEnd'];
}
export declare function KmoreFactory<D, Ctx = unknown>(options: KmoreFactoryOpts<D, Ctx>): Kmore<D, Ctx>;
export declare function createDbh(knexConfig: KnexConfig): Knex;
//# sourceMappingURL=kmore.d.ts.map