/* eslint-disable max-lines-per-function */
/* eslint-disable @typescript-eslint/no-explicit-any */
/* eslint-disable import/no-extraneous-dependencies */
import assert from 'assert';
// eslint-disable-next-line no-duplicate-imports, import/no-named-default
import { default as _knex } from 'knex';
import { KmoreBase } from './base.js';
import { createRefTables } from './builder.index.js';
import { initialConfig } from './config.js';
import { postProcessResponse, wrapIdentifier } from './helper.js';
import { createTrxProperties } from './proxy.trx.js';
import { CaseType, } from './types.js';
import { genKmoreTrxId } from './util.js';
export class Kmore extends KmoreBase {
    /**
     * Original table names, without case convertion.
     */
    refTables;
    /**
     * Create a table reference function property with camel case convertion.
     */
    camelTables;
    // readonly pascalTables: DbQueryBuilder<D, 'ref_', CaseType.pascal>
    /**
     * Create a table reference function property with snake case convertion.
     */
    snakeTables;
    /**
    * Generics parameter, do NOT access as variable!
    * Use under typescript development only.
    *
    * @example ```ts
    * const km = kmoreFactore<Db>({})
    * type DbModel = typeof km.DbModel
    * type Uid = DbModel['tb_user']['uid']  // equal to number
    * ```
    */
    DbModel;
    /**
    * Generics parameter, do NOT access as variable!
    * Use under typescript development only.
    *
    * @example ```ts
    * const km = kmoreFactore<Db>({})
    * type Dict = typeof kmore['Dict']
    * type Uid = Dict['columns']['tb_user']['alias']['tbUserUid'] // equal to literal 'tb_user.uid'
    * ```
    */
    Dict;
    postProcessResponseSet = new Set();
    trxActionOnEnd = 'rollback';
    /**
     * kmoreTrxId => trx
     */
    trxMap = new Map();
    /**
     * kmoreTrxId => Set<kmoreQueryId>
     */
    trxIdQueryMap = new Map();
    /**
     * context => Set<kmoreTrxId>
     */
    ctxTrxIdMap = new WeakMap();
    config;
    dict;
    dbId;
    dbh;
    instanceId;
    eventCallbacks;
    wrapIdentifierCaseConvert;
    constructor(options) {
        super();
        const dbId = options.dbId ? options.dbId : Date.now().toString();
        this.dbId = dbId;
        this.instanceId = options.instanceId ? options.instanceId : Symbol(`${dbId}-` + Date.now().toString());
        this.eventCallbacks = options.eventCallbacks;
        const config = {
            ...initialConfig,
            ...options.config,
        };
        config.pool = { ...initialConfig.pool, ...options.config.pool };
        this.config = config;
        // assert(options.dict, 'options.dict must be defined')
        if (options.dict) {
            // @ts-expect-error
            this.dict = options.dict;
        }
        else {
            console.info('Kmore:constructor() options.dict empty');
            // @ts-expect-error
            this.dict = void 0;
        }
        /**
         * Table identifier case convertion,
         * If not CaseType.none, will ignore value of `KnexConfig['wrapIdentifier']`
         */
        this.wrapIdentifierCaseConvert = options.wrapIdentifierCaseConvert ?? CaseType.snake;
        if (!this.config.wrapIdentifier) {
            if (this.wrapIdentifierCaseConvert === CaseType.none) {
                this.config.wrapIdentifier = defaultGlobalWrapIdentifier;
            }
            else {
                this.config.wrapIdentifier = wrapIdentifier;
            }
        }
        this.postProcessResponseSet.add(postProcessResponse);
        if (typeof this.config.postProcessResponse === 'function') {
            const fn = this.config.postProcessResponse;
            this.postProcessResponseSet.add(fn);
        }
        delete this.config.postProcessResponse;
        this.config.postProcessResponse = (result, queryContext) => this.postProcessResponse(result, queryContext);
        if (options.trxActionOnEnd) {
            this.trxActionOnEnd = options.trxActionOnEnd;
        }
        this.refTables = createRefTables(this, 'ref_', CaseType.none);
        this.camelTables = createRefTables(this, 'ref_', CaseType.camel);
        // this.pascalTables = this.createRefTables<'ref_'>('ref_', CaseType.pascal)
        this.snakeTables = createRefTables(this, 'ref_', CaseType.snake);
        this.dbh = options.dbh ? options.dbh : createDbh(this.config);
    }
    /**
     * Start a transaction.
     */
    async transaction(config) {
        const kmoreTrxId = config?.kmoreTrxId ?? genKmoreTrxId(config?.kmoreTrxId);
        delete config?.kmoreTrxId;
        const trx = await this.dbh.transaction(void 0, config);
        const trxActionOnEnd = config?.trxActionOnEnd
            ?? this.trxActionOnEnd ?? 'rollback';
        const ret = createTrxProperties({
            kmore: this,
            kmoreTrxId,
            trx,
            trxActionOnEnd,
        });
        return ret;
    }
    getTrxByKmoreQueryId(kmoreQueryId) {
        for (const [trxId, queryIdSet] of this.trxIdQueryMap.entries()) {
            if (queryIdSet.has(kmoreQueryId)) {
                const trx = this.trxMap.get(trxId);
                if (trx) {
                    return trx;
                }
            }
        }
    }
    getTrxByKmoreTrxId(id) {
        return this.trxMap.get(id);
    }
    setCtxTrxIdMap(ctx, kmoreTrxId) {
        assert(ctx, 'ctx must be defined');
        assert(kmoreTrxId, 'kmoreTrxId must be defined');
        assert(typeof ctx === 'object');
        if (!this.ctxTrxIdMap.get(ctx)) {
            this.ctxTrxIdMap.set(ctx, new Set());
        }
        this.ctxTrxIdMap.get(ctx)?.add(kmoreTrxId);
    }
    getTrxSetByCtx(ctx) {
        const ret = new Set();
        if (!ctx || typeof ctx !== 'object') {
            return ret;
        }
        const trxIdMap = this.ctxTrxIdMap.get(ctx);
        trxIdMap?.forEach((kmoreTrxId) => {
            const trx = this.trxMap.get(kmoreTrxId);
            if (trx) {
                ret.add(trx);
            }
        });
        return ret;
    }
    async finishTransaction(trx) {
        if (!trx) {
            return;
        }
        if (!trx.isCompleted()) {
            this.trxMap.delete(trx.kmoreTrxId);
        }
        switch (trx.trxActionOnEnd) {
            case 'rollback': {
                await trx.rollback();
                this.trxMap.delete(trx.kmoreTrxId);
                this.trxIdQueryMap.delete(trx.kmoreTrxId);
                break;
            }
            case 'commit': {
                await trx.commit();
                this.trxMap.delete(trx.kmoreTrxId);
                this.trxIdQueryMap.delete(trx.kmoreTrxId);
                break;
            }
            default:
                break;
        }
    }
    finishTransactionByQueryId(kmoreQueryId) {
        const trx = this.getTrxByKmoreQueryId(kmoreQueryId);
        return this.finishTransaction(trx);
    }
    destroy() {
        return this.dbh.destroy();
    }
    async raw(...args) {
        // @ts-expect-error
        return this.dbh.raw(...args);
    }
    builderClone(builder) {
        const builder2 = builder.clone();
        return builder2;
    }
    /* -------------- protected -------------- */
    postProcessResponse(result, queryContext) {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
        let ret = result;
        for (const fn of this.postProcessResponseSet) {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            ret = fn(ret, queryContext);
        }
        return ret;
    }
}
export function KmoreFactory(options) {
    const km = new Kmore(options);
    return km;
}
export function createDbh(knexConfig) {
    const inst = _knex.knex(knexConfig);
    return inst;
}
const fnKeys = [
    'CASE',
    'WHEN',
    'THEN',
    'ELSE',
    'ELS',
    'END',
];
function defaultGlobalWrapIdentifier(value, origImpl) {
    // const dem2 = '\''
    // original  IFNULL(user.distotal, 0) as foo
    // from: IFNULL(user => input: `IFNULL(user`  fix to IFNULL(user
    // from: distotal, 0) => input: `distotal, 0)` fix to distotal, 0)
    if (value.includes('(')) {
        return value;
    }
    if (value.includes(')')) {
        return value;
    }
    const upper = value.toUpperCase().trim();
    const matched = fnKeys.find(key => upper.includes(key));
    if (matched) {
        if (upper.startsWith('CASE ')) {
            return value;
        }
        if (upper.endsWith(' END')) {
            return value;
        }
    }
    const line = origImpl(value);
    if (value === '' && line === '``') {
        return '\'\'';
    }
    return line;
}
//# sourceMappingURL=kmore.js.map